'''
Functions for helping curate BSE basis set data
'''

from .printing import *
from .compare import *
from .compare_ref import *
from .metadata import *
from .read_g94 import *
from .read_turbomole import *



"""
Test for validating the formatting of the json files
"""

import glob
import os

import pytest
from basis_set_exchange import api, validator

_data_dir = api._default_data_dir

_all_files = glob.glob(os.path.join(_data_dir, '*.json'))
_all_files.extend(glob.glob(os.path.join(_data_dir, '*', '*.json')))

@pytest.mark.parametrize('file_path', _all_files)
def test_valid(file_path):
    # Validate all the data files in the data directory
    # against their respective schemas
    if file_path.endswith('METADATA.json'):
        return
    if file_path.endswith('REFERENCES.json'):
        validator.validate('references', file_path)
    elif file_path.endswith('.table.json'):
        validator.validate('table', file_path)
    elif file_path.endswith('.element.json'):
        validator.validate('element', file_path)
    else:
        validator.validate('component', file_path)



'''
Comparison of basis data against authoritative sources
'''

import copy
from .. import get_basis
from .. import fileio
from ..misc import compact_elements
from .compare import shells_difference, potentials_difference, compare_electron_shells, compare_ecp_pots
from .read_turbomole import read_turbomole


def _fix_uncontracted(basis):
    '''
    Forces the contraction coefficient of uncontracted shells to 1.0
    '''

    for el in basis['basis_set_elements'].values():
        if 'element_electron_shells' not in el:
            continue

        for sh in el['element_electron_shells']:
            if len(sh['shell_coefficients']) == 1 and len(sh['shell_coefficients'][0]) == 1:
                sh['shell_coefficients'][0][0] = '1.0000000'

            # Some uncontracted shells don't have a coefficient
            if len(sh['shell_coefficients']) == 0:
                sh['shell_coefficients'].append(['1.0000000'])

    return basis


def _print_list(lst):
    '''
    Prints a list from a comparison
    '''

    a = compact_elements(lst)
    if a is not None:
        return a
    else:
        return ""


def _compare_basis_against_ref(basis_name, src_data, version=None):
    '''
    Compares basis set data from an authoritative source against bse data
    '''
    bse_data = get_basis(basis_name, version=version)
    all_src = list(src_data['basis_set_elements'].keys())

    not_in_src = []  # Found in BSE, not in the reference
    not_in_bse = all_src.copy()  # Found in the reference, not in the BSE
    no_diff = []  # Elements for which there is no difference
    some_diff = []  # Elements that are different
    big_diff = []  # Elements that are substantially different

    for k, v in bse_data['basis_set_elements'].items():
        if k not in all_src:
            not_in_src.append(k)
            continue

        print()
        print("-------------------------------------")
        print(" Element ", k)
        src_el = src_data['basis_set_elements'][k]

        max_rdiff_el = 0.0
        max_rdiff_ecp = 0.0

        if 'element_electron_shells' in v:
            max_rdiff_el = shells_difference(v['element_electron_shells'], src_el['element_electron_shells'])
            print()
        if 'element_ecp' in v:
            max_rdiff_ecp = potentials_difference(v['element_ecp'], src_el['element_ecp'])
            v['element_ecp'] = src_el['element_ecp']
            print()

        max_rdiff = max(max_rdiff_el, max_rdiff_ecp)

        # Handle some differences
        if max_rdiff == float('inf'):
            big_diff.append(k)
        elif max_rdiff == 0.0:
            no_diff.append(k)
        else:
            some_diff.append(k)

        not_in_bse.remove(k)

    print()
    print("     Not in src: ", _print_list(not_in_src))
    print("     Not in bse: ", _print_list(not_in_bse))
    print("  No difference: ", _print_list(no_diff))
    print("Some difference: ", _print_list(some_diff))
    print(" BIG difference: ", _print_list(big_diff))
    print()


def _read_ref_data(src_filepath, file_type):
    type_readers = {'turbomole': read_turbomole}

    if file_type not in type_readers:
        raise RuntimeError("Unknown file type to read '{}'".format(file_type))

    data = type_readers[file_type](src_filepath)
    return _fix_uncontracted(data)


def compare_basis_against_ref(basis_name, src_filepath, file_type, version=None):
    '''
    Compares basis set data from an authoritative source against bse data
    '''

    src_data = _read_ref_data(src_filepath, file_type)
    src_data = _fix_uncontracted(src_data)
    _compare_basis_against_ref(basis_name, src_data, version)


def _replace_shell_data(old_shells, src_shells):
    '''
    Replaces exponents and coefficients of old_data with that in src_data
    '''

    new_shells = []
    for sh1 in src_shells:
        # Find its likely candidate in old_shells
        for sh2 in old_shells:
            if compare_electron_shells(sh1, sh2, rel_tol=0.0):
                new_shells.append(sh2)
                break
            elif compare_electron_shells(sh1, sh2, rel_tol=1e-3):
                sh_tmp = copy.deepcopy(sh2)
                sh_tmp['shell_exponents'] = sh1['shell_exponents']
                sh_tmp['shell_coefficients'] = sh1['shell_coefficients']
                new_shells.append(sh_tmp)
                break
        else:
            sh_tmp = copy.deepcopy(sh1)
            sh_tmp['data_source'] = None
            new_shells.append(sh_tmp)

    return new_shells


def _replace_ecp_data(old_pots, src_pots):
    '''
    Replaces ECP exponents and coefficients of old_data with that in src_data
    '''

    new_pots = []
    for pot1 in src_pots:
        # Find its likely candidate in old_pots
        for pot2 in old_pots:
            if compare_ecp_pots(pot1, pot2, rel_tol=0.0):
                new_pots.append(pot2)
                break
            elif compare_ecp_pots(pot1, pot2, rel_tol=1e-3):
                pot_tmp = copy.deepcopy(pot2)
                pot_tmp['potential_r_exponents'] = pot1['potential_r_exponents']
                pot_tmp['potential_gaussian_exponents'] = pot1['potential_gaussian_exponents']
                pot_tmp['potential_coefficients'] = pot1['potential_coefficients']
                new_pots.append(pot_tmp)
                break
        else:
            pot_tmp = copy.deepcopy(pot1)
            pot_tmp['data_source'] = 'MISSING.json'
            new_pots.append(pot_tmp)

    return new_pots


def replace_basis_data(basis_name, src_filepath, file_type, version=None, inplace=False):
    '''
    Replaces data in basis set files with those from a reference

    Data is replaced only if it doesn't exactly match.

    If inplace is True, then the data in the original files is overwritten.
    Otherwise, the new data is written to a file with the same name, but
    ending with .diff
    '''

    src_data = _read_ref_data(src_filepath, file_type)
    bse_data = get_basis(basis_name, version=version)

    all_src = list(src_data['basis_set_elements'].keys())

    not_in_src = []  # Found in BSE, not in the reference
    not_in_bse = all_src.copy()  # Found in the reference, not in the BSE

    replace_map = {}
    for k, v in bse_data['basis_set_elements'].items():
        if k not in all_src:
            not_in_src.append(k)
            continue
        not_in_bse.remove(k)

        src_el = src_data['basis_set_elements'][k]

        new_shells = []
        new_pots = []
        if 'element_electron_shells' in v:
            bse_shells = v['element_electron_shells']
            src_shells = src_el['element_electron_shells']
            new_shells = _replace_shell_data(bse_shells, src_shells)
        if 'element_ecp' in v:
            bse_pots = v['element_ecp']
            src_pots = src_el['element_ecp']
            new_pots = _replace_ecp_data(bse_pots, src_pots)

        # add to the replacement map
        for sh in new_shells:
            fpath = sh.pop('data_source')
            if not fpath in replace_map:
                replace_map[fpath] = {}
            if not k in replace_map[fpath]:
                replace_map[fpath][k] = {}
            if not 'element_electron_shells' in replace_map[fpath][k]:
                replace_map[fpath][k]['element_electron_shells'] = []
            replace_map[fpath][k]['element_electron_shells'].append(sh)

        for pot in new_pots:
            fpath = pot.pop('data_source')
            if not fpath in replace_map:
                replace_map[fpath] = {}
            if not k in replace_map[fpath]:
                replace_map[fpath][k] = {}
            if not 'element_ecp' in replace_map[fpath][k]:
                replace_map[fpath][k]['element_ecp'] = []
            replace_map[fpath][k]['element_ecp'].append(pot)

    # Actually do the replacements
    for fpath, replace_data in replace_map.items():

        # Read the original data from the file
        if fpath is not None:
            orig_data = fileio.read_json_basis(fpath)
        else:
            orig_data = {'basis_set_elements': {}}
            fpath = "MISSING.json"

        for el, replace_eldata in replace_data.items():
            if el in orig_data['basis_set_elements']:
                orig_eldata = orig_data['basis_set_elements'][el]

                # orig_eldata is the original data for an element
                # replace_eldata is what we are replacing with

                # Sort into the same order as orig data
                new_shell_order = []
                new_ecp_order = []
                at_shell_end = []
                at_ecp_end = []

                if 'element_electron_shells' in replace_eldata:
                    for sh1 in replace_eldata['element_electron_shells']:
                        for idx, sh2 in enumerate(orig_eldata['element_electron_shells']):
                            if compare_electron_shells(sh1, sh2, rel_tol=1e-3):
                                new_shell_order.append((idx, sh1))
                                break
                        else:
                            at_shell_end.append(sh1)

                if 'element_ecp' in replace_eldata:
                    for pot1 in replace_eldata['element_ecp']:
                        for idx, pot2 in enumerate(orig_eldata['element_ecp']):
                            if compare_ecp_pots(pot1, pot2, rel_tol=1e-3):
                                new_ecp_order.append((idx, pot1))
                                break
                        else:
                            at_ecp_end.append(pot1)

                if len(new_shell_order) > 0:
                    new_shell_order = sorted(new_shell_order)
                    new_shell_order = [x[1] for x in new_shell_order]
                    new_shell_order.extend(at_shell_end)
                    orig_eldata['element_electron_shells'] = new_shell_order

                if len(new_ecp_order) > 0:
                    new_ecp_order = sorted(new_ecp_order)
                    new_ecp_order = [x[1] for x in new_ecp_order]
                    new_ecp_order.extend(at_ecp_end)
                    orig_eldata['element_ecp'] = new_ecp_order
            else:
                orig_data['basis_set_elements'][el] = replace_eldata

        if not inplace:
            fpath = fpath + '.diff'

        fileio.write_json_basis(fpath, orig_data)

    print()
    print("     Not in src: ", _print_list(not_in_src))
    print("     Not in bse: ", _print_list(not_in_bse))
    print()

    return bse_data



'''
Conversion of basis sets to various formats
'''

from .bsejson import write_json
from .nwchem import write_nwchem
from .g94 import write_g94
from .gamess_us import write_gamess_us
from .psi4 import write_psi4

converter_map = {
    'json': {
        'display': 'JSON',
        'extension': '.json',
        'comment': None,
        'function': write_json
    },
    'nwchem': {
        'display': 'NWChem',
        'extension': '.nw',
        'comment': '#',
        'function': write_nwchem
    },
    'gaussian94': {
        'display': 'Gaussian94',
        'extension': '.gbs',
        'comment': '!',
        'function': write_g94
    },
    'gamess_us': {
        'display': 'GAMESS US',
        'extension': '.bas',
        'comment': '!',
        'function': write_gamess_us
    },
    'psi4': {
        'display': 'Psi4',
        'extension': '.gbs',
        'comment': '!',
        'function': write_psi4
    }
}



"""
Compares version 0 with historical BSE
"""

import bz2
import os
import re

import pytest
from basis_set_exchange import api

# Find the dir with all the bse files
_my_dir = os.path.dirname(os.path.abspath(__file__))
_hist_data_dir = os.path.join(_my_dir, 'basis_set_exchange-historical')
_bse_data_dir = os.path.join(_hist_data_dir, 'bse-formatted')
_hist_testfile = os.path.join(_hist_data_dir, '.is_historical')
_hist_testfile_exists = os.path.isfile(_hist_testfile)

# Skip everying in this file if the submodule isn't checked out
pytestmark = pytest.mark.skipif(_hist_testfile_exists is not True, reason="Historical basis data not downloaded")

# Load all the metadata once
_bs_metadata = api.get_metadata()
_bs_formats = list(api.get_formats().keys())
_bs_formats.remove('json') # Don't need to test json (nothing to compare to)
_true_false = [ True, False ]

# Only the names with a version zero
_bs_names_only_v0 = [x for x,y in _bs_metadata.items() if '0' in y['versions']]

# Read the mapping of new BSE names to old BSE files
with open(os.path.join(_my_dir, 'bse_v0_map.txt'), 'r') as handle:
    _bse_v0_map = dict(re.split('\t+', x.strip()) for x in handle.readlines())

# Mapping of old->new symbols    
_sym_replace_map = { 'Uun': 'Ds', 'UUN': 'DS',
                     'Uuu': 'Rg', 'UUU': 'RG',
                     'Uub': 'Cn', 'UUB': 'CN',
                     'Uut': 'Nh', 'UUT': 'NH',
                     'Uuq': 'Fl', 'UUQ': 'FL',
                     'Uup': 'Mc', 'UUP': 'MC',
                     'Uuh': 'Lv', 'UUH': 'LV',
                     'Uus': 'Ts', 'UUS': 'TS',
                     'Uuo': 'Og', 'UUO': 'OG'
                   }

# Mapping of old->new names
_name_replace_map = { 'ununnilium':  'darmstadtium', 'UNUNNILIUM':  'DARMSTADTIUM',
                      'unununium':   'roentgenium',  'UNUNUNIUM':   'ROENTGENIUM',
                      'ununbium':    'copernicium',  'UNUNBIUM':    'COPERNICIUM',
                      'ununtrium':   'nihonium',     'UNUNTRIUM':   'NIHONIUM',
                      'ununquadium': 'flerovium',    'UNUNQUADIUM': 'FLEROVIUM',
                      'ununpentium': 'moscovium',    'UNUNPENTIUM': 'MOSCOVIUM',
                      'ununhexium':  'livermorium',  'UNUNHEXIUM':  'LIVERMORIUM',
                      'ununseptium': 'tennessine',   'UNUNSEPTIUM': 'TENNESSINE',
                      'ununoctium':  'oganesson',    'UNUNOCTIUM':  'OGANESSON',
                      'elemtn113' :  'nihonium',     'ELEMTN113' :  'NIHONIUM',
                      'element115' :  'moscovium',   'ELEMENT115' : 'MOSCOVIUM',
                      'element117' :  'tennessine',  'ELEMENT117' : 'TENNESSINE'
                    }


def _process_commonline(x):
    x = x.strip()

    # collapse multiple spaces
    x = re.sub(r' +', r' ', x)

    # remove trailing zero after a decimal point
    #x = re.sub(r'\.(\d+?)0+\b', r'.\1 ', x)
    #x = re.sub(r'\.(\d+?)0+([Ee])', r'.\1\2', x)

    # Substitute old symbols (BSE had old symbols for some elements)
    for k,v in _sym_replace_map.items():
        x = x.replace(k, v)

    # Substitute names, too
    for k,v in _name_replace_map.items():
        x = x.replace(k, v)

    # Replace D with E for exponents
    x = re.sub(r'(\d)D([\+\-]?)(\d)', r'\1E\2\3', x)

    return x


def _process_nwchem(lines):
    # Process lines from an nwchem-formatted file
    lines = [ _process_commonline(x) for x in lines ]

    # Remove comments and empty lines
    lines = [ x for x in lines if not x.startswith('#') and x != '']

    return lines


def _process_gaussian(lines):
    # Process lines from an g94-formatted file
    lines = [ _process_commonline(x) for x in lines ]

    # The 'potential' lines are just comments (but have to be there)
    lines = [ re.sub(r'^([spdfghijk]|ul)[ \-].*potential$', r'', x) for x in lines ]

    # Remove comments and empty lines
    lines = [ x for x in lines if not x.startswith('!') and x != '']

    return lines


def _process_gamess_us(lines):
    # Process lines from a GAMESS US formatted file
    lines = [ _process_commonline(x) for x in lines ]

    # Remove comments and empty lines
    lines = [ x for x in lines if not x.startswith('!') and x != '']

    # The old bse misspells 'PHOSPHORUS' as 'PHOSPHOROUS'
    lines = [ x.replace('PHOSPHOROUS', 'PHOSPHORUS') for x in lines ]

    # The 'potential' parts are just comments
    lines = [ re.sub(r'-----.*-----', '', x) for x in lines ]

    return lines


def _lines_equivalent(line1, line2):
    # Bail early if they are exactly the same
    if line1 == line2:
        return True

    # Split the string by spaces, and compare
    splt1 = re.split(r' +', line1)
    splt2 = re.split(r' +', line2)

    if len(splt1) != len(splt2):
        return False

    for i in range(len(splt1)):
        # See if they are floats and do exact compare
        try:
            x1 = float(splt1[i])
            x2 = float(splt2[i])
            if x1 != x2:
                return False
        except ValueError:
            if splt1[i] != splt2[i]:
                return False

    return True

# Maps formats to processing functions and bse suffixes
_format_map = { 'nwchem' : ('NWChem', _process_nwchem),
                'gaussian94' : ('Gaussian94', _process_gaussian),
                'gamess_us' : ('GAMESS-US', _process_gamess_us)
              }


@pytest.mark.slow
@pytest.mark.parametrize('basis_name', _bs_names_only_v0)
@pytest.mark.parametrize('fmt', _bs_formats)
@pytest.mark.parametrize('opt_gen', _true_false)
def test_v0_with_bse(basis_name, fmt, opt_gen):

    basis_meta = _bs_metadata[basis_name]
    fmt_info = _format_map[fmt]

    if not basis_name in _bse_v0_map:
        raise RuntimeError("Mapping from BSE to old BSE name doesn't exist for " + basis_name)

    # Read in the data from the old BSE
    bse_name = _bse_v0_map[basis_name]
    fmt_suffix = '.' + fmt_info[0] + '.bz2'

    # See if the bse optimized-general file is there. If not, it is the
    # same as the un-optimized
    if opt_gen:
        fmt_suffix2 = '.min' + fmt_suffix
        bse_file = os.path.join(_bse_data_dir, bse_name + fmt_suffix2)
        if os.path.isfile(bse_file):
            fmt_suffix = fmt_suffix2

    bse_file = os.path.join(_bse_data_dir, bse_name + fmt_suffix)
    if not os.path.isfile(bse_file):
        raise FileNotFoundError("File {} does not exist (for comparing basis {})".format(bse_file, basis_name))

    with bz2.open(bse_file, 'rt') as f:
        bse_data = f.readlines()
    bse_data = fmt_info[1](bse_data)

    # read in data from the new bse (version 0)
    new_data = api.get_basis(basis_name, version='0', fmt=fmt, optimize_general=opt_gen, header=False)
    new_data = new_data.split('\n')
    new_data = fmt_info[1](new_data)

    if len(new_data) != len(bse_data):
        print("BSE FILE: " + bse_file)
        #open('T.new', 'w').write("\n".join(new_data))
        #open('T.old', 'w').write("\n".join(bse_data))
        raise RuntimeError("Basis set: {} different number of lines: {} vs {}".format(basis_name, len(new_data), len(bse_data)))

    for i in range(len(new_data)):
        new_line = new_data[i]
        bse_line = bse_data[i]

        if not _lines_equivalent(new_line, bse_line):
            errstr = '''Difference found. Line {}
                        New: {}
                        Old: {}'''.format(i, new_line, bse_line)
            #open('T.new', 'w').write("\n".join(new_data))
            #open('T.old', 'w').write("\n".join(bse_data))
            raise RuntimeError(errstr)




'''
Conversion of references to various formats
'''

from .bib import write_bib
from .txt import write_txt
from .bsejson import write_json

converter_map = {
    'json': {
        'display': 'JSON',
        'extension': '.json',
        'comment': None,
        'function': write_json
    },
    'bib': {
        'display': 'BibTeX',
        'extension': '.bib',
        'comment': '%',
        'function': write_bib
    },
    'txt': {
        'display': 'Plain Text',
        'extension': '.txt',
        'comment': None,
        'function': write_txt
    }
}



'''
Helper functions for writing out basis set in various formats
'''


def _find_point(x):
    if isinstance(x, int):
        return 0
    else:
        return x.index('.')


def _determine_leftpad(column, point_place):
    '''Find how many spaces to put before a column of numbers
       so that all the decimal points line up

    This function takes a column of decimal numbers, and returns a
    vector containing the number of spaces to place before each number
    so that (when possible) the decimal points line up.


    Parameters
    ----------
    column : list
        Numbers that will be printed as a column
    point_place : int
        Number of the character column to put the decimal point
    '''

    # Find the number of digits before the decimal
    ndigits_left = [_find_point(x) for x in column]

    # find the padding per entry, filtering negative numbers
    return [max((point_place - 1) - x, 0) for x in ndigits_left]


def write_matrix(mat, point_place, convert_exp=False):

    # Padding for the whole matrix
    pad = [_determine_leftpad(c, point_place[i]) for i, c in enumerate(mat)]

    # Use the transposes (easier to write out by row)
    pad = list(map(list, zip(*pad)))
    mat = list(map(list, zip(*mat)))

    lines = ''
    for r, row in enumerate(mat):
        line = ''
        for c, val in enumerate(row):
            sp = pad[r][c] - len(line)
            # ensure at least one space
            sp = max(sp, 1)
            line += ' ' * sp + str(mat[r][c])
        lines += line + '\n'

    if convert_exp is True:
        lines = lines.replace('e', 'D')
        lines = lines.replace('E', 'D')

    return lines



"""
Some helper functions related to handling of references/citations
"""

import textwrap


def compact_references(basis_dict, ref_data):
    """
    Creates a mapping of elements to reference keys

    A list is returned, with each element being a dictionary
    with entries 'reference_info' containing data for (possibly) multiple references,
    and 'elements' which is a list of element Z numbers
    that those references apply to

    Parameters
    ----------
    basis_dict : dict
        Dictionary containing basis set information
    ref_data : dict
        Dictionary containing all reference information
    """

    element_refs = []

    # Create a mapping of elements -> reference information
    for el, eldata in basis_dict['basis_set_elements'].items():

        # elref is a list of dict
        # dict is { 'reference_description': str, 'reference_keys': [keys] }
        elref = eldata['element_references']

        for x in element_refs:
            if x['reference_info'] == elref:
                x['elements'].append(el)
                break
        else:
            element_refs.append({'reference_info': elref, 'elements': [el]})

    for item in element_refs:

        # Loop over a list of dictionaries for this group of elements
        # Note that reference_data needs to be in the same order as the reference_keys
        for elref in item['reference_info']:
            elref['reference_data'] = {k: ref_data[k] for k in elref['reference_keys']}

    return element_refs


def reference_text(ref):
    '''Convert a single reference to plain text format

    Parameters
    ----------
    ref : dict
        Information about a single reference
    '''

    ref_wrap = textwrap.TextWrapper(initial_indent='', subsequent_indent=' ' * 8)

    s = u''
    if ref['type'] == 'article':
        s += ref_wrap.fill(u', '.join(ref['authors'])) + '\n'
        s += ref_wrap.fill(ref['title']) + '\n'
        s += u'{}, {}, {} ({})'.format(ref['journal'], ref['volume'], ref['page'], ref['year'])
        s += u'\n' + ref['doi']
    elif ref['type'] == 'incollection':
        s += ref_wrap.fill(u', '.join(ref['authors']))
        s += ref_wrap.fill(u'\n{}'.format(ref['title']))
        s += ref_wrap.fill(u'\nin \'{}\''.format(ref['booktitle']))
        if 'editors' in ref:
            s += ref_wrap.fill(u'\ned. ' + u', '.join(ref['editors']))
        if 'series' in ref:
            s += u'\n{}, {}, {} ({})'.format(ref['series'], ref['volume'], ref['page'], ref['year'])
        if 'doi' in ref:
            s += u'\n' + ref['doi']
    elif ref['type'] == 'techreport':
        s += ref_wrap.fill(u', '.join(ref['authors']))
        s += ref_wrap.fill(u'\n{}'.format(ref['title']))
        s += u'\n\'{}\''.format(ref['institution'])
        s += u'\nTechnical Report {}'.format(ref['number'])
        s += u'\n{}'.format(ref['year'])
        if 'doi' in ref:
            s += u'\n' + ref['doi']
    else:
        raise RuntimeError('Cannot handle reference type {}'.format(ref['type']))
    return s



'''
Helpers for printing pieces of basis sets
'''

from .. import lut
from .. import manip
from ..converters.common import write_matrix


def print_electron_shell(shell, shellidx=None):
    '''Print the data for an electron shell

    If shellidx (index of the shell) is not None, it will also be printed
    '''
    am = shell['shell_angular_momentum']
    amchar = lut.amint_to_char(am)
    amchar = amchar.upper()

    if shellidx is not None:
        shellidx = ''

    exponents = shell['shell_exponents']
    coefficients = shell['shell_coefficients']
    ncol = len(coefficients) + 1

    point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol + 1)]
    print("Shell {} '{}': AM {} ({})".format(shellidx, shell['shell_region'], am, amchar))
    print(write_matrix([exponents, *coefficients], point_places))


def print_ecp_pot(pot):
    '''Print the data for an ECP potential
    '''

    am = pot['potential_angular_momentum']
    amchar = lut.amint_to_char(am)

    rexponents = pot['potential_r_exponents']
    gexponents = pot['potential_gaussian_exponents']
    coefficients = pot['potential_coefficients']

    point_places = [0, 10, 33]
    print('Potential: {} potential'.format(amchar))
    print(write_matrix([rexponents, gexponents, *coefficients], point_places))


def print_element(z, eldata, print_references=True):
    '''Print all data for an element

    Parameters
    ----------
    z : integer
        Element Z-number
    eldata:
        Data for the element to be printed
    '''

    sym = lut.element_sym_from_Z(z, True)

    print()
    print('Element: {}   Contraction: {}'.format(sym, manip.contraction_string(eldata)))
    if print_references:
        if 'element_references' in eldata:
            print('References:')
            for x in eldata['element_references']:
                print('    ' + ', '.join(x['reference_keys']))
                print('        ' + x['reference_description'])
        else:
            print('References: NONE')

    if 'element_electron_shells' in eldata:
        for shellidx, shell in enumerate(eldata['element_electron_shells']):
            print_electron_shell(shell, shellidx)

    if 'element_ecp' in eldata:
        print('ECP: Element: {}   Number of electrons: {}'.format(sym, eldata['element_ecp_electrons']))

        # Sort lowest->highest, then put the highest at the beginning
        ecp_list = sorted(eldata['element_ecp'], key=lambda x: x['potential_angular_momentum'])
        ecp_list.insert(0, ecp_list.pop())

        for pot in ecp_list:
            print_ecp_pot(pot)


def print_component_basis(basis, elements=None):
    '''Print a component basis set

    If elements is not None, only the specified elements will be printed
    (list of integers)
    '''
    print("Description: " + basis['basis_set_description'])
    print('References: ' + ', '.join(basis['basis_set_references']))

    eldata = basis['basis_set_elements']

    # Filter to the given elements
    if elements is None:
        elements = list(eldata.keys())
    else:
        elements = [k for k in eldata.keys() if k in elements]

    # Electron Basis
    for z in elements:
        print_element(z, eldata[z], False)


def print_element_basis(basis, elements=None):
    '''Print an element basis set

    If elements is not None, only the specified elements will be printed
    (list of integers)
    '''
    print("Basis set: " + basis['basis_set_name'])
    print("Description: " + basis['basis_set_description'])

    eldata = basis['basis_set_elements']

    if elements is None:
        elements = list(eldata.keys())
    else:
        elements = [k for k in eldata.keys() if k in elements]

    # strings
    complist = {z: eldata[z]['element_components'] for z in elements}

    # Header line
    print('{:4} {}'.format("El", "Components"))
    print('-' * 80)
    for z in elements:
        sym = lut.element_sym_from_Z(z, True)
        print('{:4} {}'.format(sym, complist[z][0]))
        for v in complist[z][1:]:
            print('{:4} {}'.format('', v))
    print()


def print_table_basis(basis, elements=None):
    '''Print a full table basis set

    If elements is not None, only the specified elements will be printed
    (list of integers)
    '''

    print("Basis set: " + basis['basis_set_name'])
    print("Description: " + basis['basis_set_description'])
    print("Role: " + basis['basis_set_role'])
    print()

    eldata = basis['basis_set_elements']

    if elements is None:
        elements = list(eldata.keys())
    else:
        elements = [k for k in eldata.keys() if k in elements]

    # strings
    complist = {z: eldata[z]['element_entry'] for z in elements}

    # Header line
    print('{:4} {}'.format("El", "Entry"))
    print('-' * 80)
    for z in elements:
        sym = lut.element_sym_from_Z(z, True)
        print('{:4} {}'.format(sym, complist[z]))

    print()



'''
Functions for looking up element names, numbers, and symbols
and for converting between angular momentum conventions

This module has functions for looking up element information by
symbol, name, or number. It also has functions for converting
angular momentum between integers (0, 1, 2) and letters (s, p, d).
'''

# Contains the symbols, names, and Z numbers for all known elements
# NOTE: Some Z number have multiple entries. This is to allow for querying
#       based on older, systematic names/symbols (like uuh), however a query
#       of the Z number should return the 'new', official name (livermorium or lv)
# yapf: disable
_data_table = [
    ("h",     1, "hydrogen"),     ("he",    2, "helium"),        ("li",    3, "lithium"),
    ("be",    4, "beryllium"),    ("b",     5, "boron"),         ("c",     6, "carbon"),
    ("n",     7, "nitrogen"),     ("o",     8, "oxygen"),        ("f",     9, "fluorine"),
    ("ne",   10, "neon"),         ("na",   11, "sodium"),        ("mg",   12, "magnesium"),
    ("al",   13, "aluminum"),     ("si",   14, "silicon"),       ("p",    15, "phosphorus"),
    ("s",    16, "sulfur"),       ("cl",   17, "chlorine"),      ("ar",   18, "argon"),
    ("k",    19, "potassium"),    ("ca",   20, "calcium"),       ("sc",   21, "scandium"),
    ("ti",   22, "titanium"),     ("v",    23, "vanadium"),      ("cr",   24, "chromium"),
    ("mn",   25, "manganese"),    ("fe",   26, "iron"),          ("co",   27, "cobalt"),
    ("ni",   28, "nickel"),       ("cu",   29, "copper"),        ("zn",   30, "zinc"),
    ("ga",   31, "gallium"),      ("ge",   32, "germanium"),     ("as",   33, "arsenic"),
    ("se",   34, "selenium"),     ("br",   35, "bromine"),       ("kr",   36, "krypton"),
    ("rb",   37, "rubidium"),     ("sr",   38, "strontium"),     ("y",    39, "yttrium"),
    ("zr",   40, "zirconium"),    ("nb",   41, "niobium"),       ("mo",   42, "molybdenum"),
    ("tc",   43, "technetium"),   ("ru",   44, "ruthenium"),     ("rh",   45, "rhodium"),
    ("pd",   46, "palladium"),    ("ag",   47, "silver"),        ("cd",   48, "cadmium"),
    ("in",   49, "indium"),       ("sn",   50, "tin"),           ("sb",   51, "antimony"),
    ("te",   52, "tellurium"),    ("i",    53, "iodine"),        ("xe",   54, "xenon"),
    ("cs",   55, "cesium"),       ("ba",   56, "barium"),        ("la",   57, "lanthanum"),
    ("ce",   58, "cerium"),       ("pr",   59, "praseodymium"),  ("nd",   60, "neodymium"),
    ("pm",   61, "promethium"),   ("sm",   62, "samarium"),      ("eu",   63, "europium"),
    ("gd",   64, "gadolinium"),   ("tb",   65, "terbium"),       ("dy",   66, "dysprosium"),
    ("ho",   67, "holmium"),      ("er",   68, "erbium"),        ("tm",   69, "thulium"),
    ("yb",   70, "ytterbium"),    ("lu",   71, "lutetium"),      ("hf",   72, "hafnium"),
    ("ta",   73, "tantalum"),     ("w",    74, "tungsten"),      ("re",   75, "rhenium"),
    ("os",   76, "osmium"),       ("ir",   77, "iridium"),       ("pt",   78, "platinum"),
    ("au",   79, "gold"),         ("hg",   80, "mercury"),       ("tl",   81, "thallium"),
    ("pb",   82, "lead"),         ("bi",   83, "bismuth"),       ("po",   84, "polonium"),
    ("at",   85, "astatine"),     ("rn",   86, "radon"),         ("fr",   87, "francium"),
    ("ra",   88, "radium"),       ("ac",   89, "actinium"),      ("th",   90, "thorium"),
    ("pa",   91, "protactinium"), ("u",    92, "uranium"),       ("np",   93, "neptunium"),
    ("pu",   94, "plutonium"),    ("am",   95, "americium"),     ("cm",   96, "curium"),
    ("bk",   97, "berkelium"),    ("cf",   98, "californium"),   ("es",   99, "einsteinium"),
    ("fm",  100, "fermium"),      ("md",  101, "mendelevium"),   ("no",  102, "nobelium"),
    ("lr",  103, "lawrencium"),   ("rf",  104, "rutherfordium"), ("db",  105, "dubnium"),
    ("sg",  106, "seaborgium"),   ("bh",  107, "bohrium"),       ("hs",  108, "hassium"),
    ("mt",  109, "meitnerium"),   ("uun", 110, "ununnilium"),    ("ds",  110, "darmstadtium"),
    ("uuu", 111, "unununium"),    ("rg",  111, "roentgenium"),   ("uub", 112, "ununbium"),
    ("cn",  112, "copernicium"),  ("uut", 113, "ununtrium"),     ("nh",  113, "nihonium"),
    ("uuq", 114, "ununquadium"),  ("fl",  114, "flerovium"),     ("uup", 115, "ununpentium"),
    ("mc",  115, "moscovium"),    ("uuh", 116, "ununhexium"),    ("lv",  116, "livermorium"),
    ("uus", 117, "ununseptium"),  ("ts",  117, "tennessine"),    ("uuo", 118, "ununoctium"),
    ("og",  118, "oganesson")
]
# yapf: enable

# Maps Z to element data
# Note that only the last entry for a particular Z number will
# be stored (which contains the new, official data rather than
# the older systematic names for some elements)
_element_Z_map = {x[1]: x for x in _data_table}

# Maps element symbol to element data
_element_sym_map = {x[0]: x for x in _data_table}

# Maps element name to element data
_element_name_map = {x[2]: x for x in _data_table}

# Maps AM characters to integers (the integer is the
# index of the character in this string)
# This is the 'hik' convention, where AM=7 is k
_amchar_map_hik = 'spdfghiklmnoqrtuvwxyzabce'

# This is the 'hij' convention, where AM=7 is j
_amchar_map_hij = 'spdfghijklmnoqrtuvwxyzabce'


def element_data_from_Z(Z):
    '''Obtain elemental data given a Z number

    An exception is thrown if the Z number is not found
    '''

    # Z may be a str
    if isinstance(Z, str) and Z.isdecimal():
        Z = int(Z)

    if Z not in _element_Z_map:
        raise KeyError('No element data for Z = {}'.format(Z))
    return _element_Z_map[Z]


def element_data_from_sym(sym):
    '''Obtain elemental data given an elemental symbol

    The given symbol is not case sensitive

    An exception is thrown if the symbol is not found
    '''

    sym_lower = sym.lower()
    if sym_lower not in _element_sym_map:
        raise KeyError('No element data for symbol \'{}\''.format(sym))
    return _element_sym_map[sym_lower]


def element_data_from_name(name):
    '''Obtain elemental data given an elemental name

    The given name is not case sensitive

    An exception is thrown if the name is not found
    '''

    name_lower = name.lower()
    if name_lower not in _element_name_map:
        raise KeyError('No element data for name \'{}\''.format(name))
    return _element_name_map[name_lower]


def element_name_from_Z(Z, normalize=False):
    '''Obtain an element's name from its Z number

    An exception is thrown if the Z number is not found

    If normalize is True, the first letter will be capitalized
    '''

    r = element_data_from_Z(Z)[2]
    if normalize:
        return normalize_element_name(r)
    else:
        return r


def element_sym_from_Z(Z, normalize=False):
    '''Obtain an element's symbol from its Z number

    An exception is thrown if the Z number is not found

    If normalize is True, the first letter will be capitalized
    '''

    r = element_data_from_Z(Z)[0]
    if normalize:
        return normalize_element_symbol(r)
    else:
        return r


def element_Z_from_sym(sym):
    '''Obtain an element's Z-number given its symbol

    An exception is thrown if the symbol is not found
    '''

    return element_data_from_sym(sym)[1]


def normalize_element_symbol(sym):
    '''Normalize the capitalization of an element symbol (where only the first letter is capitalized

    For example, converts "he" to "He" and "uUo" to "Uuo"
    '''

    # We can cheat, since this will be the same algorithm as normalizing the name
    return normalize_element_name(sym)


def normalize_element_name(name):
    '''Normalize the capitalization of an element name

    For example, converts "helium" to "Helium" and "aRgOn" to "Argon"
    '''

    name = name.lower()
    name2 = name[0].upper()
    name2 += name[1:]
    return name2


def amint_to_char(am, hij=False, use_L=False):
    '''Convert an angular momentum integer to a character

    The input is a list (to handle sp, spd, ... orbitals). The return
    value is a string

    For example, converts [0] to 's' and [0,1,2] to 'spd'

    If hij is True, the ordering spdfghijkl is used. Otherwise, the
    ordering will be spdfghikl (skipping j)

    If use_L is True, sp shells ([0,1]) will return l instead
    '''

    if use_L and am == [0, 1]:
        return 'l'
    if hij:
        amchar_map = _amchar_map_hij
    else:
        amchar_map = _amchar_map_hik

    amchar = []

    for a in am:
        if a < 0:
            raise IndexError('Angular momentum must be a positive integer (not {})'.format(a))
        if a >= len(amchar_map):
            raise IndexError('Angular momentum {} out of range. Must be less than {}'.format(a, len(amchar_map)))
        amchar.append(amchar_map[a])

    return ''.join(amchar)


def amchar_to_int(amchar, hij=False):
    '''Convert an angular momentum integer to a character

    The return value is a list of integers (to handle sp, spd, ... orbitals)

    For example, converts 'p' to [1] and 'sp' to [0,1]

    If hij is True, the ordering spdfghijkl is used. Otherwise, the
    ordering will be spdfghikl (skipping j)
    '''

    if hij:
        amchar_map = _amchar_map_hij
    else:
        amchar_map = _amchar_map_hik

    amchar_lower = amchar.lower()

    amint = []

    for c in amchar_lower:
        if c not in amchar_map:
            raise KeyError('Angular momentum character {} is not valid'.format(c))

        amint.append(amchar_map.index(c))

    return amint



"""
Tests for reference handling
"""

import glob
import json
import os

import pytest
from basis_set_exchange import api, refconverters, misc

_data_dir = api._default_data_dir

# _all_files shouldn't contain .table. files
_all_files = glob.glob(os.path.join(_data_dir, '*', '*.json'))
_all_component_files = [x for x in _all_files if '.element.' not in x]


@pytest.mark.parametrize("elements, expected", 
                         [ ([1], "H"),
                           ([1,2], "H,He"),
                           ([1,10], "H,Ne"),
                           ([1,2,3,11,23,24], "H-Li,Na,V,Cr")])
def test_compact_string(elements, expected):
    assert misc.compact_elements(elements) == expected



@pytest.mark.parametrize('file_path', _all_component_files)
def test_filenames(file_path):
    with open(file_path, 'r') as f:
        file_refs = json.load(f)['basis_set_references'] 
        if len(file_refs) > 0:
            ref_str = '_'.join(file_refs)
        else:
            ref_str = 'noref'
        ref_str = '_' + ref_str

        # Strip off '.0.json', '.1.json', etc
        base_name = os.path.splitext(file_path)[0]
        base_name = os.path.splitext(base_name)[0]

        # Base name should end with the reference string
        assert base_name.endswith(ref_str)



'''
Conversion of basis sets to Gaussian format
'''

from .. import lut
from .. import manip
from .common import write_matrix


def write_gamess_us(basis):
    '''Converts a basis set to GAMESS-US
    '''

    s = ''

    basis = manip.uncontract_general(basis)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['basis_set_elements'].items() if 'element_electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['basis_set_elements'].items() if 'element_ecp' in v]

    # Electron Basis
    if len(electron_elements) > 0:
        # electronic part starts with $DATA
        s += '$DATA\n'

        for z in electron_elements:
            data = basis['basis_set_elements'][z]

            el_name = lut.element_name_from_Z(z).upper()
            s += '\n' + el_name + "\n"

            for shell in data['element_electron_shells']:
                exponents = shell['shell_exponents']
                coefficients = shell['shell_coefficients']
                ncol = len(coefficients) + 2  #include index column
                nprim = len(exponents)

                am = shell['shell_angular_momentum']
                amchar = lut.amint_to_char(am, hij=True, use_L=True).upper()
                s += '{}   {}\n'.format(amchar, nprim)

                # 1-based indexing
                idx_column = list(range(1, nprim + 1))
                point_places = [0] + [4 + 8 * i + 15 * (i - 1) for i in range(1, ncol)]
                s += write_matrix([idx_column, exponents, *coefficients], point_places)

        s += "$END"

    # Write out ECP
    if len(ecp_elements) > 0:
        s += "\n\n$ECP\n"

        for z in ecp_elements:
            data = basis['basis_set_elements'][z]
            sym = lut.element_sym_from_Z(z).upper()
            max_ecp_am = max([x['potential_angular_momentum'][0] for x in data['element_ecp']])
            max_ecp_amchar = lut.amint_to_char([max_ecp_am], hij=True)

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['element_ecp'], key=lambda x: x['potential_angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '{}-ECP GEN    {}    {}\n'.format(sym, data['element_ecp_electrons'], max_ecp_am)

            for pot in ecp_list:
                rexponents = pot['potential_r_exponents']
                gexponents = pot['potential_gaussian_exponents']
                coefficients = pot['potential_coefficients']
                nprim = len(rexponents)

                am = pot['potential_angular_momentum']
                amchar = lut.amint_to_char(am, hij=False)

                # Title line
                if am[0] == max_ecp_am:
                    s += '{:<5} ----- {}-ul potential -----\n'.format(nprim, amchar)
                else:
                    s += '{:<5} ----- {}-{} potential -----\n'.format(nprim, amchar, max_ecp_amchar)

                point_places = [8, 23, 32]
                s += write_matrix([*coefficients, rexponents, gexponents], point_places)

        s += "$END\n"

    return s



'''
Helpers for handling BSE metadata
'''

import os
from collections import OrderedDict

from .. import api
from .. import compose
from .. import fileio
from .. import misc


def create_metadata_file(output_path, data_dir):
    '''Creates a METADATA.json file from a data directory

    The file is written to output_path
    '''

    basis_filelist = fileio.get_basis_filelist(data_dir)

    metadata = {}
    for bs_file_relpath in basis_filelist:
        # Obtain the version from the filename
        filebase = os.path.splitext(bs_file_relpath)[0]  # remove .json
        filebase = os.path.splitext(filebase)[0]  # remove .table
        filebase, ver = os.path.splitext(filebase)  # remove .[version]
        ver = ver[1:]  # Remove the period

        # Fully compose the basis set from components
        bs = compose.compose_table_basis(bs_file_relpath, data_dir)

        # Prepare the metadata
        tr_name = misc.transform_basis_name(bs['basis_set_name'])
        display_name = bs['basis_set_name']
        defined_elements = sorted(list(bs['basis_set_elements'].keys()), key=lambda x: int(x))
        description = bs['basis_set_description']
        revision_desc = bs['basis_set_revision_description']
        role = bs['basis_set_role']
        family = bs['basis_set_family']
        auxiliaries = bs['basis_set_auxiliaries']

        function_types = set()
        for e in bs['basis_set_elements'].values():
            if 'element_electron_shells' in e:
                for s in e['element_electron_shells']:
                    function_types.add(s['shell_function_type'])
            if 'element_ecp' in e:
                function_types.add('ecp')

        function_types = sorted(list(function_types))

        single_meta = OrderedDict(
            [('display_name', display_name), ('file_relpath', bs_file_relpath), ('family', family),
             ('description', description), ('revdesc', revision_desc), ('role', role), ('auxiliaries', auxiliaries),
             ('functiontypes', function_types), ('elements', defined_elements)])

        if not tr_name in metadata:
            metadata[tr_name] = {'versions': {ver: single_meta}}
        else:
            metadata[tr_name]['versions'][ver] = single_meta

    # sort the versions and find the max version
    # Also place display_name, role, auxiliaries into the top level for this basis
    for k, v in metadata.items():
        latest = max(v['versions'].keys())
        latest_data = v['versions'][latest]
        metadata[k] = OrderedDict([('display_name', latest_data['display_name']), ('latest_version',
                                                                                   latest), ('family',
                                                                                             latest_data['family']),
                                   ('role', latest_data['role']), ('functiontypes', latest_data['functiontypes']),
                                   ('auxiliaries', latest_data['auxiliaries']), ('versions',
                                                                                 OrderedDict(
                                                                                     sorted(v['versions'].items())))])

    # Remove these from under versions
    to_remove = ['display_name', 'role', 'auxiliaries', 'family', 'functiontypes']
    for v in metadata.values():
        for ver in v['versions'].values():
            for x in to_remove:
                ver.pop(x)

    # Write out the metadata
    metadata = OrderedDict(sorted(list(metadata.items())))
    fileio._write_plain_json(output_path, metadata)



import os
from .. import lut


def read_g94(basis_path):
    '''Reads a G94-formatted file and converts it to a dictionary with the
       usual BSE fields

       Note that the gaussian format does not store all the fields we
       have, so some fields are left blank
    '''

    skipchars = '!'

    if not os.path.isfile(basis_path):
        raise RuntimeError('G94 basis set path \'{}\' does not exist'.format(basis_path))

    fname = os.path.basename(basis_path)

    with open(basis_path, 'r') as f:
        basis_lines = [l.strip() for l in f]
        basis_lines = [l for l in basis_lines if l and not l[0] in skipchars]

    bs_data = {
        'molssi_bse_schema': {
            'schema_type': 'component',
            'schema_version': '0.1'
        },
        'basis_set_description': fname,
        'basis_set_references': [],
        'basis_set_elements': {}
    }

    i = 0
    if basis_lines[0] == '****':
        i += 1  # skip initial ****

    while i < len(basis_lines):
        line = basis_lines[i]
        elementsym = line.split()[0]

        # Some gaussian files have a dash before the element
        if elementsym[0] == '-':
            elementsym = elementsym[1:]

        element_Z = lut.element_Z_from_sym(elementsym)
        element_Z = str(element_Z)

        element_data = {'element_electron_shells': []}

        i += 1
        if "ECP" in basis_lines[i]:
            raise NotImplementedError("Parsing of ECP not implemented")
            #maxam = int(lsplt[1])
            #n_elec = int(lsplt[2])

            #i += 1
            #for j in range(maxam + 1):
            #    shell_am = lut.amchar_to_int(basis_lines[i][0])[0]
            #    i += 1
            #    n_entries = int(basis_lines[i])
            #    i += 1

            #    ecp_shell = {'angularMomentum': shell_am}
            #    ecp_exponents = []
            #    ecp_coefficients = []

            #    for k in range(n_entries):
            #        lsplt = basis_lines[i].split()
            #        exp_exponents.append(lsplt[0])
            #        exp_coefficients.append(lsplt[1])
            #        i += 1

            #    bs_data['elements'][element_Z]['ecp'].append(ecp_shell)
        else:
            while basis_lines[i] != '****':
                lsplt = basis_lines[i].split()
                shell_am = lut.amchar_to_int(lsplt[0])
                nprim = int(lsplt[1])

                shell = {
                    'shell_function_type': 'gto',
                    'shell_harmonic_type': 'spherical',
                    'shell_region': 'valence',
                    'shell_angular_momentum': shell_am
                }

                exponents = []
                coefficients = []

                i += 1
                for j in range(nprim):
                    line = basis_lines[i].replace('D', 'E')
                    line = line.replace('d', 'E')
                    lsplt = line.split()
                    exponents.append(lsplt[0])
                    coefficients.append(lsplt[1:])
                    i += 1

                shell['shell_exponents'] = exponents

                # We need to transpose the coefficient matrix
                # (we store a matrix with primitives being the column index and
                # general contraction being the row index)
                shell['shell_coefficients'] = list(map(list, zip(*coefficients)))

                element_data['element_electron_shells'].append(shell)
                bs_data['basis_set_elements'][element_Z] = element_data

        i += 1

    return bs_data



"""
Validation of bibtex output
"""

import pytest
from basis_set_exchange import api

from . import check_bibtex

# Load all the metadata once
_bs_names = api.get_all_basis_names()

@pytest.mark.slow
@pytest.mark.skipif(check_bibtex.available is False, reason="Latex/Bibtex commands not found, so I can't verify bib files")
@pytest.mark.parametrize('basis_name', _bs_names)
def test_bibtex(basis_name):
    bib_str = api.get_references(basis_name, fmt='bib')
    check_bibtex.validate_bibtex(bib_str)



"""
Tests for BSE metadata
"""

import json
import os

from basis_set_exchange import api, curate

_data_dir = api._default_data_dir

def test_get_metadata():
    api.get_metadata()


def test_metadata_uptodate():
    old_metadata = os.path.join(_data_dir, 'METADATA.json')
    new_metadata = os.path.join(_data_dir, 'METADATA.json.new')
    curate.create_metadata_file(new_metadata, _data_dir)

    with open(old_metadata, 'r') as f:
        old_data = json.load(f)
    with open(new_metadata, 'r') as f:
        new_data = json.load(f)

    os.remove(new_metadata)

    assert old_data == new_data



"""
Tests for the BSE main API
"""

import random

import basis_set_exchange as bse
import pytest
from basis_set_exchange import lut

from .common_testvars import *

# Use random for getting sets of elements
random.seed(rand_seed, version=2)

@pytest.mark.parametrize('basis_name', bs_names)
def test_get_basis_1(basis_name):
    """For all versions of basis sets, test a simple get_basis
    """
    this_metadata = bs_metadata[basis_name]
    for ver in this_metadata['versions'].keys():
        bse.get_basis(basis_name, version=ver)


@pytest.mark.parametrize('basis_name', bs_names)
def test_get_basis_2(basis_name):
    """For all versions of basis sets, test a simple get_basis
       with different element selections
    """
    this_metadata = bs_metadata[basis_name]
    latest = this_metadata['latest_version']
    avail_elements = this_metadata['versions'][latest]['elements']
    nelements = random.randint(1, len(avail_elements))
    selected_elements = random.sample(avail_elements, nelements)

    # Change some selected elements to strings 
    for idx in range(len(selected_elements)):
        if idx % 3 == 1:
            selected_elements[idx] = lut.element_sym_from_Z(selected_elements[idx])
        elif idx % 3 == 2:
            selected_elements[idx] = str(selected_elements[idx])

    bs = bse.get_basis(basis_name, elements=selected_elements)
    assert len(bs['basis_set_elements']) == len(selected_elements)


@pytest.mark.parametrize('basis_name', bs_names_sample)
@pytest.mark.parametrize('bool_opts', bool_matrix(5))
def test_get_basis_3(basis_name, bool_opts):
    """For a sample of basis sets, test different options
    """
    bse.get_basis(basis_name,
                  uncontract_general=bool_opts[0],
                  uncontract_segmented=bool_opts[1],
                  uncontract_spdf=bool_opts[2],
                  make_general=bool_opts[3],
                  optimize_general=bool_opts[4])


@pytest.mark.parametrize('basis_name', bs_names_sample)
@pytest.mark.parametrize('fmt', bs_formats)
def test_get_basis_4(basis_name, fmt):
    """For a sample of basis sets, test getting different formats
       of the latest version
    """
    bse.get_basis(basis_name, fmt=fmt)


@pytest.mark.parametrize('basis_name', bs_names_sample)
def test_get_basis_memo(basis_name):
    """For a sample of basis sets, test memoization
    """
    bs1 = bse.get_basis(basis_name)
    bs2 = bse.get_basis(basis_name)

    # Should be equal, but not aliased
    assert bs1 == bs2
    assert bs1['basis_set_elements'] is not bs2['basis_set_elements']


@pytest.mark.parametrize('basis_name', bs_names)
@pytest.mark.parametrize('fmt', ref_formats)
def test_get_references_1(basis_name, fmt):
    """ Tests getting references for all basis sets
    """
    this_metadata = bs_metadata[basis_name]
    for ver in this_metadata['versions'].keys():
        bse.get_references(basis_name, fmt=fmt, version=ver)

        avail_elements = this_metadata['versions'][ver]['elements']
        nelements = random.randint(1, len(avail_elements))
        selected_elements = random.sample(avail_elements, nelements)
        bse.get_references(basis_name, elements=selected_elements, fmt=fmt, version=ver)


@pytest.mark.parametrize('primary_basis,role,expected', role_tests)
def test_lookup_by_role(primary_basis, role, expected):
    """Test looking up data by role
    """
    bs = bse.lookup_basis_by_role(primary_basis, role)
    assert bs.lower() == expected.lower()


@pytest.mark.parametrize('basis_name', bs_names)
def test_notes(basis_name):
    """Test getting family, family notes, and basis set notes
    """
    bse.get_basis_notes(basis_name)
    fam = bse.get_basis_family(basis_name)
    bse.get_family_notes(fam)

